# -*- coding: utf8 -*-
from __future__ import unicode_literals

import time
import logging

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    Segment,
    HeaderSegment,
    HprimString as HL7String,
    HprimDate as HL7Date,
    HprimForeign as HL7Foreign,
)

logger = logging.getLogger("driverng.mpl")


class MplHl7SegmentMSH(HeaderSegment):
    sending_application = HL7String(3, default="MCA")
    receiving_application = HL7String(5, default="MPL")
    message_date = HL7Date(7)
    message_type = HL7String((9, 1), default="OUL")
    message_code = HL7String((9, 2), default="R22")
    message_label = HL7String((9, 3), default="OUL_R22")
    control_id = HL7String(10)
    processing_id = HL7String(11, default="P")
    version = HL7String(12, default="2.5")
    accept_acknowledgment_type = HL7String(15, default="NE")
    application_acknowlegment_type = HL7String(16, default="AL")

    patients = HL7Foreign("PID", "header")
    specimens = HL7Foreign("SPM", "header")


class MplHl7SegmentPID(Segment):
    mpl_patient_id = HL7String(3)
    # Admission number in MPL
    admission_number = HL7String(4)
    patient_name = HL7String((6, 1))
    patient_firstname = HL7String((6, 2))
    maiden_name = HL7String(7)
    birthdate = HL7Date(8, fmt_date="%Y%m%d")
    sex = HL7String(9)

    visits = HL7Foreign("PV1", "patient")
    specimens = HL7Foreign("SPM", "patient")
    infos = HL7Foreign("NTE", "patient")


class MplHl7SegmentPV1(Segment):
    location = HL7String(4)
    attending_doctor = HL7String(8)

    orders = HL7Foreign("OBR", "patient")
    specimens = HL7Foreign("SPM", "patient")
    infos = HL7Foreign("NTE", "patient")


class MplHl7SegmentSPM(Segment):
    specimen_id = HL7String((3, 1))
    control_code = HL7String((3, 2))
    specimen_type = HL7String(4)

    orders = HL7Foreign("OBR", "patient")
    infos_samples = HL7Foreign("SAC", "patient")
    common_order = HL7Foreign("ORC", "patient")
    infos = HL7Foreign("NTE", "patient")


class MplHl7SegmentSAC(Segment):
    barcode_number = HL7String(4)
    rack = HL7String(11)
    position = HL7String(12)

    orders = HL7Foreign("OBR", "patient")
    infos = HL7Foreign("NTE", "patient")


class MplHl7SegmentOBR(Segment):
    barcode = HL7String(4)
    test_code = HL7String((5, 1))
    name_test = HL7String((5, 2))
    code_doctor = HL7String((17, 1))
    name_doctor = HL7String((17, 2))
    collection_date = HL7Date((28, 4))
    # S = Stat
    # R = Routine
    sample_priority = HL7String((28, 6))

    common_order = HL7Foreign("ORC", "orders")
    results = HL7Foreign("OBX", "orders")
    infos = HL7Foreign("NTE", "patient")


class MplHl7SegmentOBX(Segment):
    value_type = HL7String(3)
    test_code = HL7String(4)
    observation_result = HL7String(6)
    units = HL7String(7)
    abnormal_flags = HL7String(9)
    result_status = HL7String(12)
    result_date = HL7Date(15)
    observer = HL7String(17)
    method = HL7String(18)
    instrument = HL7String(19)

    dilution = HL7Foreign("TCD", "results")
    infos = HL7Foreign("NTE", "patient")


class MplHl7SegmentTCD(Segment):
    test_code = HL7String(2)
    num1 = HL7String((3, 2), default="1")
    suffix = HL7String((3, 3))
    dilution = HL7String((3, 4))


class MplHl7SegmentORC(Segment):
    order_control = HL7String(2)
    order_id = HL7String(4)
    barcode = HL7String(5)
    order_status = HL7String(6)


class MplHl7SegmentNTE(Segment):
    source_comment = HL7String(3)
    comment = HL7String(4)


class MplHl7SegmentINV(Segment):
    reagent_code = HL7String(2)
    bottle_type = HL7String(3)
    reagent_type = HL7String(4)
    reagent_serial_number = HL7String(5)
    reagent_lot = HL7String(17)


class MplHl7SegmentEQU(Segment):
    instrument_identifier = HL7String((2, 1))
    instrument_name = HL7String((2, 2))
    event_date = HL7Date(3)
    instrument_status = HL7String(4)


class Driver(AstmMessageV2):
    SPEC = {
        "code": "mpl_hl7",
        "label": "MPL HL7",
    }

    SEGMENT_CLASS_MAP = {
        "MSH": MplHl7SegmentMSH,
        "PID": MplHl7SegmentPID,
        "PV1": MplHl7SegmentPV1,
        "SPM": MplHl7SegmentSPM,
        "SAC": MplHl7SegmentSAC,
        "OBR": MplHl7SegmentOBR,
        "OBX": MplHl7SegmentOBX,
        "TCD": MplHl7SegmentTCD,
        "ORC": MplHl7SegmentORC,
        "NTE": MplHl7SegmentNTE,
        "INV": MplHl7SegmentINV,
        "EQU": MplHl7SegmentEQU,
    }
    DEFAULT_SEPARATORS = "|^~\\&"
    FIRST_SEP_INDEX = 3
    SEGMENT_ENDING = "\r"

    OPTIONS = [
        AstmMessageV2.Option("mpl_hl7_sender", label="Code émetteur", default="MCA"),
        AstmMessageV2.Option("mpl_hl7_receiver", label="Code récepteur", default="MPL"),
    ]

    def __init__(self, **kwargs):
        self.sender_code = kwargs.pop("mpl_hl7_sender", "MCA")
        self.receiver_code = kwargs.pop("mpl_hl7_receiver", "MPL")
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        return MplHl7SegmentMSH(type_segment="MSH", **kwargs)

    def to_string(self, message, separators=None, **kwargs):
        return super(Driver, self).to_string(message, separators, **kwargs) + "\r"

    def encode_worklist(self, worklist, context):
        # Encode un message de type résultat SIL
        if worklist.context != "result":
            raise NotImplementedError("context %s" % (worklist.context,))
        worklist_sample_types = list(set([s.type for s in worklist.samples]))
        qc = worklist_sample_types == ["control"]
        message = self.create(
            sending_application=self.sender_code,
            receiving_application=self.receiver_code,
            message_date=context["now"],
            control_id=context.get(
                "control_id", int(time.time() * 1000)),
        )
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            priority={"S": "S", "A": "S", worklist.ELSE: "R"},
            action={"add": "NW", "add": "XO", "del": worklist.IGNORE},
        ):
            if not qc:
                seg_pid = message.patients.add(
                    patient_id=patient.patient_id,
                    patient_identifier_list=patient.patient_id,
                    patient_name=patient.last_name,
                    patient_firstname=patient.first_name,
                    maiden_name=patient.maiden_name,
                    birthdate=patient.birthdate,
                    sex=patient.sex,
                )
                physician = [p for p in patient.physicians if p.type == "attending"]
                seg_pv1 = seg_pid.visits.add(
                    location=patient.location,
                    attending_doctor=physician[0].code if physician else "",
                )
                for sample, action, priority in patient.samples:
                    prescriber = [p for p in patient.physicians if p.type == "prescriber"]
                    seg_spm = seg_pv1.specimens.add(
                        specimen_id=sample.sample_id,
                        specimen_type=sample.nature["code"],
                    )
                    for result in sample.results:
                        seg_obr = seg_spm.orders.add(
                            barcode=sample.sample_id,
                            test_code=result.code,
                            name_test=result.code,
                            code_doctor=prescriber[0].code if prescriber else "",
                            name_doctor=prescriber[0].code if prescriber else "",
                            collection_date=sample.collection_date,
                            sample_priority=priority,
                        )
                        seg_obx = seg_obr.results.add(
                            value_type=result.type,
                            test_code=result.code,
                            observation_result=result.value,
                            units=result.unit,
                            result_status="F",
                            result_date=result.date,
                            observer=result.validator,
                            instrument=result.instrument,
                        )
                        if result.dilution:
                            seg_obx.dilution.add(
                                test_code=result.code,
                                suffix=seg_obx.dilution,
                            )
        return message

    def decode_message(self, message):
        # Decode un message de type demande SIL
        worklist = Worklist(
            context="order",
            sender_id=message.sending_application,
            receiver_id=message.receiving_application,
        )
        for seg_pid in message.patients:
            location = None
            if seg_pid.visits:
                if seg_pid.visits[0].location:
                    location = "MPL{}".format(seg_pid.visits[0].location)
                patient_specimens = seg_pid.visits[0].specimens
            else:
                patient_specimens = seg_pid.specimens
            for seg_spm in patient_specimens:
                for seg_obr in seg_spm.orders:
                    if seg_pid.admission_number:
                        request_id = seg_pid.admission_number
                    else:
                        request_id = seg_obr.barcode
                    sample = worklist.get_or_create(
                        type="patient",
                        sample_id=seg_obr.barcode,
                        request_id=request_id,
                        patient_id=seg_pid.mpl_patient_id,
                        last_name=seg_pid.patient_name,
                        first_name=seg_pid.patient_firstname,
                        maiden_name=seg_pid.maiden_name,
                        birthdate=seg_pid.birthdate,
                        sex=seg_pid.sex,
                        location=location,
                        collection_date=seg_obr.collection_date,
                    )
                    action = "add"
                    if seg_obr.common_order:
                        code_action = seg_obr.common_order[0].order_control
                        if code_action == "NW":
                            action = "add"
                        elif code_action == "CA":
                            action = "del"
                        else:
                            action = "add"
                    sample.orders.get_or_create(
                        code=seg_obr.test_code,
                        priority=seg_obr.sample_priority,
                        action=action,
                    )
                    if seg_obr.code_doctor:
                        sample.physicians.get_or_create(
                            code="MPL{}".format(seg_obr.code_doctor),
                            name=seg_obr.name_doctor,
                            type="prescriber",
                        )
                    for seg_obx in seg_obr.results:
                        sample.results.get_or_create(
                            type=seg_obx.value_type,
                            code=seg_obx.test_code,
                            value=seg_obx.observation_result,
                            date=seg_obx.result_date,
                        )
        return worklist
