# -*- coding: utf8 -*-
from __future__ import unicode_literals

import io
import os
import time
import paramiko
from paramiko.dsskey import DSSKey
from paramiko.ecdsakey import ECDSAKey
from paramiko.ed25519key import Ed25519Key
from paramiko.rsakey import RSAKey


from driverng.transport.base import (
    BaseFileTransport,
    BaseFileBackend,
    TransportException,
)


def catch_sftp_exception(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (
            paramiko.SSHException,
            paramiko.ssh_exception.AuthenticationException,
            paramiko.ssh_exception.BadAuthenticationType,
            paramiko.ssh_exception.BadHostKeyException,
            paramiko.ssh_exception.ChannelException,
            paramiko.ssh_exception.NoValidConnectionsError,
            paramiko.ssh_exception.PartialAuthentication,
            paramiko.ssh_exception.PasswordRequiredException,
            paramiko.ssh_exception.ProxyCommandFailure,
        ) as e:
            raise TransportException(e)

    return wrapper


class TLS(object):
    def __init__(
        self, certificate_key=None, certificate_key_password=None, ssh_keytype=None
    ):
        self.certificate_key = certificate_key
        self.certificate_key_password = certificate_key_password
        self.ssh_keytype = ssh_keytype


class SFTP(BaseFileBackend):
    def __init__(
        self,
        hostname=None,
        port=None,
        username=None,
        password=None,
        del_command_disable=False,
        **kwargs
    ):
        self.hostname = hostname
        self.username = username
        self.password = password
        self.port = port
        self._ttl = time.time()
        self._sftp = None
        self.del_command_disable = del_command_disable
        self.logger = kwargs.get("logger")
        self.pkey = None
        self.tls = TLS(
            kwargs.pop("certificate_key", None),
            kwargs.pop("certificate_key_password", None),
            kwargs.pop("ssh_keytype", None),
        )
        self._client = paramiko.SSHClient()
        self.logger.info("load host keys")
        self._client.load_system_host_keys()
        self._client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    def load_certs(self):
        if not self.tls:
            return
        if self.tls.certificate_key and self.tls.ssh_keytype:
            key_mapping = {
                "ed25519": Ed25519Key,
                "rsa": RSAKey,
                "ecdsa": ECDSAKey,
                "dsa": DSSKey,
            }
            pkeyklass = key_mapping.get(self.tls.ssh_keytype)
            if not pkeyklass:
                self.logger.error(
                    "unknown ssh key type: {}".format(self.tls.ssh_keytype)
                )
                return
            with io.StringIO() as pkf:
                pkf.write(self.tls.certificate_key)
                pkf.seek(0)
                self.pkey = pkeyklass.from_private_key(
                    pkf,
                    password=self.tls.certificate_key_password
                    if self.tls.certificate_key_password
                    else "",
                )

    @property
    def _expired(self):
        return (time.time() - self._ttl) > 120

    @catch_sftp_exception
    def _reinit_connection(self):
        if self._sftp is not None:
            self._sftp.close()
            self._client.close()
        self.load_certs()
        self._client.connect(
            hostname=self.hostname,
            port=self.port,
            username=self.username,
            password=self.password,
            passphrase=self.tls.certificate_key_password
            if self.tls.certificate_key_password
            else "",
            pkey=self.pkey,
            look_for_keys=False,
            allow_agent=False,
            compress=True,
        )
        self.logger.debug("login {} (sftp)".format(self.username))
        self._sftp = self._client.open_sftp()
        self.logger.info(
            "Connected to sftp://%s@%s:%s", self.username, self.hostname, self.port
        )

    @property
    def sftp(self):
        if self._sftp is None or self._expired:
            self._reinit_connection()
        self._ttl = time.time()
        return self._sftp

    @catch_sftp_exception
    def cwd(self, path):
        self.sftp.cwd(path)

    @catch_sftp_exception
    def list(self, path=None):
        files = []
        for filepath in self.sftp.listdir(path):
            files.append(os.path.basename(filepath))
        return files

    @catch_sftp_exception
    def put(self, data, filename):
        return self.sftp.putfo(io.BytesIO(data), filename)

    @catch_sftp_exception
    def get(self, filename):
        buf = io.BytesIO()
        self.sftp.getfo(filename, buf)
        return buf.getvalue()

    @catch_sftp_exception
    def delete(self, filename):
        if not self.del_command_disable:
            return self.sftp.unlink(filename)


class SFTPTransport(BaseFileTransport):
    def __init__(
        self,
        hostname=None,
        username=None,
        port=None,
        password=None,
        host=None,
        **kwargs
    ):
        del_command_disable = kwargs.pop("del_command_disable", False)
        if host is not None and hostname is None and port is None:
            if ":" in host:
                hostname, port = host.rsplit(":", 1)
                port = int(port)
            else:
                hostname = host
        backend = SFTP(
            hostname or "localhost",
            port or 22,
            username or "anonymous",
            password,
            del_command_disable,
            **kwargs
        )
        kwargs.pop("certificate_key", None),
        kwargs.pop("certificate_key_password", None),
        kwargs.pop("ssh_keytype", None),
        super(SFTPTransport, self).__init__(backend, **kwargs)
